# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math

from metadrive.third_party.kitsunetsuki.base.armature import is_left_bone, is_bone_matches
from metadrive.third_party.kitsunetsuki.base.objects import get_parent


class ArmatureMixin(object):
    def _make_vrm_bone(self, gltf_node_id, bone):
        vrm_bone = {
            'bone': None,
            'node': gltf_node_id,
            'useDefaultValues': True,
            'extras': {
                'name': bone.name,
            }
        }

        def is_hips(bone):
            return is_bone_matches(bone, ('hips', ))

        def is_upper_leg(bone, strict=True):
            names = ['thigh']
            if not strict:
                names.append('leg')
            is_upper = is_bone_matches(bone, names)
            is_child = is_hips(get_parent(bone))
            return is_upper or is_child

        def is_lower_leg(bone):
            is_lower = is_bone_matches(bone, ('calf', 'shin', 'knee'))
            is_child = is_upper_leg(get_parent(bone), strict=False)
            return is_lower or is_child

        def is_hand(bone):
            return is_bone_matches(bone, ('hand', 'wrist'))

        side = 'left' if is_left_bone(bone) else 'right'

        parents = []
        for i in range(1, 3 + 1):
            parent = get_parent(bone, i)
            if parent:
                parents.append(parent)

        if is_hips(bone):
            vrm_bone['bone'] = 'hips'

        elif (is_bone_matches(bone, ('upperchest', ))
              or (is_bone_matches(bone, ('spine', )) and is_hips(get_parent(bone, 3)))):
            vrm_bone['bone'] = 'upperChest'

        elif (is_bone_matches(bone, ('chest', )) or (is_bone_matches(bone,
                                                                     ('spine', )) and is_hips(get_parent(bone, 2)))):
            vrm_bone['bone'] = 'chest'

        elif is_bone_matches(bone, ('spine', )):
            vrm_bone['bone'] = 'spine'

        elif is_bone_matches(bone, ('neck', )):
            vrm_bone['bone'] = 'neck'

        elif is_bone_matches(bone, ('head', )):
            vrm_bone['bone'] = 'head'

        elif is_bone_matches(bone, ('eye', )):
            vrm_bone['bone'] = '{}Eye'.format(side)

        elif is_bone_matches(bone, ('foot', 'ankle')):
            vrm_bone['bone'] = '{}Foot'.format(side)

        elif is_lower_leg(bone):
            vrm_bone['bone'] = '{}LowerLeg'.format(side)

        elif is_upper_leg(bone):
            vrm_bone['bone'] = '{}UpperLeg'.format(side)

        elif is_bone_matches(bone, ('toe', )):
            vrm_bone['bone'] = '{}Toes'.format(side)

        elif is_bone_matches(bone, ('shoulder', 'clavicle')):
            vrm_bone['bone'] = '{}Shoulder'.format(side)

        elif is_bone_matches(bone, ('lowerarm', 'lower_arm', 'forearm', 'elbow')):
            vrm_bone['bone'] = '{}LowerArm'.format(side)

        elif is_bone_matches(bone, ('upperarm', 'upper_arm', 'arm')):
            vrm_bone['bone'] = '{}UpperArm'.format(side)

        elif any(map(is_hand, parents)):  # hand in parents -> finger
            if is_hand(get_parent(bone, 3)):  # 3 level deep parent
                part_name = 'Distal'
            elif is_hand(get_parent(bone, 2)):  # 2 level deep parent
                part_name = 'Intermediate'
            else:  # 1 level deep parent - direct parent
                part_name = 'Proximal'

            if is_bone_matches(bone, ('thumb', )):
                vrm_bone['bone'] = '{}Thumb{}'.format(side, part_name)

            elif is_bone_matches(bone, ('index', )):
                vrm_bone['bone'] = '{}Index{}'.format(side, part_name)

            elif is_bone_matches(bone, ('middle', )):
                vrm_bone['bone'] = '{}Middle{}'.format(side, part_name)

            elif is_bone_matches(bone, ('ring', )):
                vrm_bone['bone'] = '{}Ring{}'.format(side, part_name)

            elif is_bone_matches(bone, ('pinky', 'little')):
                vrm_bone['bone'] = '{}Little{}'.format(side, part_name)

        elif is_hand(bone):
            vrm_bone['bone'] = '{}Hand'.format(side)

        return vrm_bone

    def _make_vrm_spring(self, gltf_node_id, bone):
        vrm_spring = {
            'comment': bone.name,
            'stiffiness': 1,  # The resilience of the swaying object (the power of returning to the initial pose)
            'gravityPower': 0,
            'dragForce': 0,  # The resistance (deceleration) of automatic animation
            'gravityDir': {
                'x': 0,
                'y': -1,
                'z': 0,
            },  # down
            'center': -1,
            'hitRadius': 0,
            'bones': [gltf_node_id],
            # 'colliderGroups': [],
        }

        if bone.get('jiggle_stiffness', None) is not None:
            vrm_spring['stiffiness'] = bone.get('jiggle_stiffness')

        if bone.get('jiggle_gravity', None) is not None:
            vrm_spring['gravityPower'] = bone.get('jiggle_gravity')

        if bone.get('jiggle_amplitude', None) is not None:
            max_amp = 200
            jiggle_amplitude = min(max_amp, bone.get('jiggle_amplitude'))
            vrm_spring['dragForce'] = (max_amp - jiggle_amplitude) / max_amp

        return vrm_spring

    def make_armature(self, parent_node, armature):
        gltf_armature = super().make_armature(parent_node, armature)

        vrm_bones = set()
        vrm_springs = set()
        for bone_name, bone in armature.data.bones.items():
            gltf_node_id = None
            for gltf_node_id, gltf_node in enumerate(self._root['nodes']):
                if gltf_node['name'] == bone_name:
                    break
            else:
                continue

            vrm_bone = self._make_vrm_bone(gltf_node_id, bone)

            if vrm_bone['bone'] and vrm_bone['bone'] not in vrm_bones:
                vrm_bones.add(vrm_bone['bone'])
                self._root['extensions']['VRM']['humanoid']['humanBones'].append(vrm_bone)

                fp = self._root['extensions']['VRM']['firstPerson']

                if vrm_bone['bone'] == 'head':
                    fp['firstPersonBone'] = gltf_node_id
                    fp['extras'] = {'name': bone.name}

                elif vrm_bone['bone'] == 'leftEye':
                    fp.update(
                        {
                            'lookAtHorizontalOuter': {
                                'curve': [0, 0, 0, 1, 1, 1, 1, 0],
                                'xRange': 90,
                                'yRange': 10,
                            },
                            'lookAtHorizontalInner': {
                                'curve': [0, 0, 0, 1, 1, 1, 1, 0],
                                'xRange': 90,
                                'yRange': 10,
                            },
                            'lookAtVerticalDown': {
                                'curve': [0, 0, 0, 1, 1, 1, 1, 0],
                                'xRange': 90,
                                'yRange': 10,
                            },
                            'lookAtVerticalUp': {
                                'curve': [0, 0, 0, 1, 1, 1, 1, 0],
                                'xRange': 90,
                                'yRange': 10,
                            },
                        }
                    )

                    pose_bone = armature.pose.bones[bone_name]
                    for c in pose_bone.constraints:
                        if c.type == 'LIMIT_ROTATION':
                            fp['lookAtHorizontalOuter']['xRange'] = -math.degrees(c.min_x)
                            fp['lookAtHorizontalInner']['xRange'] = math.degrees(c.max_x)
                            fp['lookAtVerticalDown']['yRange'] = -math.degrees(c.min_z)
                            fp['lookAtVerticalUp']['yRange'] = math.degrees(c.max_z)
                            break

            pose_bone = armature.pose.bones[bone_name]

            # Wiggle Bones addon
            # https://blenderartists.org/t/wiggle-bones-a-jiggle-bone-implementation-for-2-8/1154726
            if pose_bone.get('jiggle_enable', False):
                # search for root bone
                while (pose_bone.parent and pose_bone.parent.get('jiggle_enable', False)):
                    pose_bone = pose_bone.parent

                if pose_bone.name not in vrm_springs:
                    vrm_spring = self._make_vrm_spring(gltf_node_id, pose_bone)
                    vrm_springs.add(pose_bone.name)
                    self._root['extensions']['VRM']['secondaryAnimation']['boneGroups'].append(vrm_spring)

        gltf_secondary = {
            'name': 'secondary',
            'translation': [0, 0, 0],
            'rotation': [0, 0, 0, 1],
            'scale': [1, 1, 1],
        }
        self._add_child(gltf_armature, gltf_secondary)

        return gltf_armature



from metadrive.envs.metadrive_env import MetaDriveEnv, METADRIVE_DEFAULT_CONFIG
from metadrive.engine.base_engine import BaseEngine
from metadrive.engine.engine_utils import get_global_config, initialize_global_config
from metadrive.utils import Config, recursive_equal


def test_config_unchangeable():
    c = Config({"aaa": 100}, unchangeable=True)
    try:
        c['aaa'] = 1000
    except ValueError as e:
        print('Great! ', e)
    assert c['aaa'] == 100


def test_config_sync():
    """
    The config in BaseEngine should be the same as env.config, if BaseEngine exists in process
    """
    try:
        env = MetaDriveEnv({"vehicle_config": dict(show_lidar=False, show_navi_mark=False)})
        # assert recursive_equal(env.config, env.engine.global_config)
        assert recursive_equal(env.config, get_global_config())
        assert recursive_equal(env.config, BaseEngine.global_config)
        assert env.config is BaseEngine.global_config
        env.reset()
        assert env.config is env.engine.global_config is BaseEngine.global_config
        assert recursive_equal(env.config, env.engine.global_config)
        assert recursive_equal(env.config, get_global_config())
        assert recursive_equal(env.config, BaseEngine.global_config)
        env.config.update({"vehicle_config": dict(show_lidar=True, show_navi_mark=True)})
        assert recursive_equal(env.config, env.engine.global_config)
        assert recursive_equal(env.config, get_global_config())
        assert recursive_equal(env.config, BaseEngine.global_config)
        env.close()
        # assert recursive_equal(env.config, get_global_config())
        # assert recursive_equal(env.config, BaseEngine.global_config)
        env.reset()
        assert recursive_equal(env.config, env.engine.global_config)
        assert recursive_equal(env.config, get_global_config())
        assert recursive_equal(env.config, BaseEngine.global_config)
        env.engine.global_config.update({"vehicle_config": dict(show_lidar=False, show_navi_mark=False)})
        assert recursive_equal(env.config, env.engine.global_config)
        assert recursive_equal(env.config, get_global_config())
        assert recursive_equal(env.config, BaseEngine.global_config)

        camera_shape = (128, 876)
        env.config["vehicle_config"]["rgb_camera"] = camera_shape
        assert recursive_equal(env.config, env.engine.global_config)
        assert recursive_equal(env.config, get_global_config())
        assert recursive_equal(env.config, BaseEngine.global_config)
        assert env.engine.global_config["vehicle_config"]["rgb_camera"] == camera_shape
        env.reset()
        assert env.engine.global_config["vehicle_config"]["rgb_camera"] == camera_shape
        assert recursive_equal(env.config, env.engine.global_config)
        assert recursive_equal(env.config, get_global_config())
        assert recursive_equal(env.config, BaseEngine.global_config)

    finally:
        env.close()


def test_config_set_unchange():
    """
    The config in BaseEngine should be the same as env.config, if BaseEngine exists in process
    """
    try:
        env = MetaDriveEnv({"vehicle_config": dict(show_lidar=False, show_navi_mark=False)})
        evn_cfg = env.config
        assert evn_cfg is BaseEngine.global_config
        initialize_global_config(None)
        env = None
        env = MetaDriveEnv({"vehicle_config": dict(show_lidar=False, show_navi_mark=False)})
        assert env.config is not evn_cfg
        env_cfg = env.config
        initialize_global_config(None)
        assert BaseEngine.global_config is None
        env.reset()
        assert env.config is env_cfg
        assert env_cfg is BaseEngine.global_config
        env.close()
        assert env.config is env_cfg
        assert BaseEngine.global_config is None
        env.reset()
        assert env.config is env_cfg
        assert BaseEngine.global_config is env_cfg
        env.close()

        env = MetaDriveEnv({"vehicle_config": dict(show_lidar=False, show_navi_mark=False)})
        assert env.config is not env_cfg
        assert BaseEngine.global_config is env.config
        # initialize_global_config(None)
        env.reset()

        old_cfg = BaseEngine.global_config
        env.close()
        assert env.config is old_cfg
        env.reset()
        old_cfg_2 = BaseEngine.global_config
        assert old_cfg is old_cfg_2 is env.config
        env.close()
    finally:
        env.close()


def test_config_two_env():
    """
    The config in BaseEngine should be the same as env.config, if BaseEngine exists in process
    """
    cfg_1 = dict(show_lidar=False, show_navi_mark=False)
    env_1 = MetaDriveEnv({"vehicle_config": cfg_1})
    assert env_1.config is BaseEngine.global_config
    cfg_2 = dict(show_lidar=False, show_navi_mark=True)
    env_2 = MetaDriveEnv({"vehicle_config": cfg_2})
    try:
        assert get_global_config() is BaseEngine.global_config is env_2.config
        env_1.reset()
        assert get_global_config() is BaseEngine.global_config is env_1.config
        env_1.close()
        assert get_global_config() is None and BaseEngine.global_config is None
        env_2.reset()
        assert env_1.config is not env_2.config and (env_2.config is BaseEngine.global_config is get_global_config())
        env_2.close()
        assert get_global_config() is None and BaseEngine.global_config is None
    finally:
        env_1.close()
        env_2.close()


def test_config_overwrite():
    """
    The config in BaseEngine should be the same as env.config, if BaseEngine exists in process
    """
    METADRIVE_DEFAULT_CONFIG["map"] = "S"
    env = MetaDriveEnv({"vehicle_config": dict(show_lidar=False, show_navi_mark=False)})
    try:
        env.reset()
        assert env.current_map.blocks[-1].ID == "S" and len(env.current_map.blocks) == 2
        env.close()
        env = MetaDriveEnv({"map": "C", "vehicle_config": dict(show_lidar=False, show_navi_mark=False)})
        env.reset()
        assert env.current_map.blocks[-1].ID == "C" and len(env.current_map.blocks) == 2
        env.close()
    finally:
        env.close()


if __name__ == '__main__':
    test_config_sync()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'main_window.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1000, 680)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        MainWindow.setMinimumSize(QtCore.QSize(1000, 680))
        MainWindow.setMaximumSize(QtCore.QSize(1000, 680))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icon/res/icon.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        MainWindow.setWindowIcon(icon)
        MainWindow.setStyleSheet(
            "QMainWindow { background: #fff;}\n"
            "*, QLabel { font-family: Roboto; font-weight: 300; }\n"
            "\n"
            "QScrollBar {\n"
            "    background: #eee;\n"
            "}\n"
            "\n"
            "\n"
            "QScrollBar:vertical {\n"
            "                        width: 9px;\n"
            "    margin: 0;\n"
            "                      }\n"
            "\n"
            "                      QScrollBar::handle:vertical {\n"
            "                        min-height: 15px;\n"
            "                        background: #aaa;\n"
            "\n"
            "                      }\n"
            "\n"
            "\n"
            "                      QScrollBar::handle:vertical:hover {\n"
            "                        \n"
            "                        background: #999;\n"
            "\n"
            "                      }\n"
            "                      QScrollBar::add-line:vertical {\n"
            "                      }\n"
            "\n"
            "                      QScrollBar::sub-line:vertical {\n"
            "                      }\n"
            "\n"
            "\n"
            "                      QScrollBar::add-page:vertical {\n"
            "                        background: #ddd;\n"
            "\n"
            "                      }\n"
            "                      QScrollBar::sub-page:vertical {\n"
            "                        background: #ddd;\n"
            "                    }\n"
            ""
        )
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.lst_plugins = QtWidgets.QListWidget(self.centralwidget)
        self.lst_plugins.setGeometry(QtCore.QRect(20, 80, 251, 531))
        self.lst_plugins.setStyleSheet(
            "QListWidget::item {\n"
            "padding: 8px 7px 5px;\n"
            "background: #ddd;\n"
            "outline: 0 !important;\n"
            "margin-bottom: 1px;\n"
            "margin-right: 10px;\n"
            "color: #777;\n"
            "font-weight: 500;\n"
            "border: 0;\n"
            "border-radius: 0;\n"
            "}\n"
            "\n"
            "QListWidget::item:hover {\n"
            "background: #ccc;\n"
            "}\n"
            "\n"
            "QListWidget::item:selected {\n"
            "background: #555;\n"
            "color: #eee;\n"
            "}\n"
            "\n"
            "QListWidget {\n"
            "padding: 0;\n"
            "color: #eee;\n"
            "background: transparent;\n"
            "/*\n"
            "background: #222;\n"
            "padding: 5px 8px;\n"
            "border: 1px solid #444;\n"
            "*/\n"
            "}\n"
            "\n"
            "* {\n"
            "outline: 0;\n"
            "}\n"
            "\n"
            ""
        )
        self.lst_plugins.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.lst_plugins.setLineWidth(0)
        self.lst_plugins.setObjectName("lst_plugins")
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.lst_plugins.addItem(item)
        self.frame_details = QtWidgets.QFrame(self.centralwidget)
        self.frame_details.setGeometry(QtCore.QRect(281, 60, 721, 621))
        palette = QtGui.QPalette()
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(229, 229, 229))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(63, 63, 63))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Light, brush)
        brush = QtGui.QBrush(QtGui.QColor(52, 52, 52))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Midlight, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Dark, brush)
        brush = QtGui.QBrush(QtGui.QColor(28, 28, 28))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Mid, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.BrightText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(229, 229, 229))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(229, 229, 229))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Shadow, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.AlternateBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 220))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ToolTipBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ToolTipText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(229, 229, 229))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(63, 63, 63))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Light, brush)
        brush = QtGui.QBrush(QtGui.QColor(52, 52, 52))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Midlight, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Dark, brush)
        brush = QtGui.QBrush(QtGui.QColor(28, 28, 28))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Mid, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.BrightText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(229, 229, 229))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(229, 229, 229))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Shadow, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.AlternateBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 220))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ToolTipBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ToolTipText, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(229, 229, 229))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(63, 63, 63))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Light, brush)
        brush = QtGui.QBrush(QtGui.QColor(52, 52, 52))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Midlight, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Dark, brush)
        brush = QtGui.QBrush(QtGui.QColor(28, 28, 28))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Mid, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.BrightText, brush)
        brush = QtGui.QBrush(QtGui.QColor(21, 21, 21))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(229, 229, 229))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(229, 229, 229))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Shadow, brush)
        brush = QtGui.QBrush(QtGui.QColor(42, 42, 42))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.AlternateBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 220))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ToolTipBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ToolTipText, brush)
        self.frame_details.setPalette(palette)
        self.frame_details.setAutoFillBackground(False)
        self.frame_details.setStyleSheet("#frame_details { background: #e5e5e5; }")
        self.frame_details.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_details.setFrameShadow(QtWidgets.QFrame.Plain)
        self.frame_details.setLineWidth(1)
        self.frame_details.setObjectName("frame_details")
        self.lbl_plugin_version = QtWidgets.QLabel(self.frame_details)
        self.lbl_plugin_version.setGeometry(QtCore.QRect(20, 40, 391, 21))
        palette = QtGui.QPalette()
        brush = QtGui.QBrush(QtGui.QColor(255, 151, 25))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(236, 236, 236))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 151, 25))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 151, 25))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(236, 236, 236))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(236, 236, 236))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 151, 25))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(236, 236, 236))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 151, 25))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 151, 25))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(236, 236, 236))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(236, 236, 236))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 151, 25))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(236, 236, 236))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 151, 25))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 151, 25))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(236, 236, 236))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(236, 236, 236))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Window, brush)
        self.lbl_plugin_version.setPalette(palette)
        font = QtGui.QFont()
        font.setFamily("Roboto")
        font.setBold(True)
        font.setWeight(75)
        self.lbl_plugin_version.setFont(font)
        self.lbl_plugin_version.setStyleSheet("color: #ff9719; border: 0; border-radius: 4px; font-weight: bold;")
        self.lbl_plugin_version.setTextFormat(QtCore.Qt.PlainText)
        self.lbl_plugin_version.setAlignment(QtCore.Qt.AlignLeading | QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
        self.lbl_plugin_version.setObjectName("lbl_plugin_version")
        self.lbl_plugin_desc = QtWidgets.QLabel(self.frame_details)
        self.lbl_plugin_desc.setGeometry(QtCore.QRect(20, 60, 441, 51))
        self.lbl_plugin_desc.setStyleSheet("color: #777; border: 0; font-size: 11px;")
        self.lbl_plugin_desc.setTextFormat(QtCore.Qt.PlainText)
        self.lbl_plugin_desc.setAlignment(QtCore.Qt.AlignLeading | QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop)
        self.lbl_plugin_desc.setWordWrap(True)
        self.lbl_plugin_desc.setObjectName("lbl_plugin_desc")
        self.table_plugin_settings = QtWidgets.QTableWidget(self.frame_details)
        self.table_plugin_settings.setGeometry(QtCore.QRect(20, 110, 681, 491))
        palette = QtGui.QPalette()
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(240, 240, 240))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Light, brush)
        brush = QtGui.QBrush(QtGui.QColor(247, 247, 247))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Midlight, brush)
        brush = QtGui.QBrush(QtGui.QColor(120, 120, 120))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Dark, brush)
        brush = QtGui.QBrush(QtGui.QColor(160, 160, 160))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Mid, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.BrightText, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(240, 240, 240))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Shadow, brush)
        brush = QtGui.QBrush(QtGui.QColor(247, 247, 247))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.AlternateBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 220))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ToolTipBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ToolTipText, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(240, 240, 240))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Light, brush)
        brush = QtGui.QBrush(QtGui.QColor(247, 247, 247))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Midlight, brush)
        brush = QtGui.QBrush(QtGui.QColor(120, 120, 120))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Dark, brush)
        brush = QtGui.QBrush(QtGui.QColor(160, 160, 160))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Mid, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.BrightText, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(240, 240, 240))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Shadow, brush)
        brush = QtGui.QBrush(QtGui.QColor(247, 247, 247))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.AlternateBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 220))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ToolTipBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ToolTipText, brush)
        brush = QtGui.QBrush(QtGui.QColor(120, 120, 120))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(240, 240, 240))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Light, brush)
        brush = QtGui.QBrush(QtGui.QColor(247, 247, 247))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Midlight, brush)
        brush = QtGui.QBrush(QtGui.QColor(120, 120, 120))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Dark, brush)
        brush = QtGui.QBrush(QtGui.QColor(160, 160, 160))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Mid, brush)
        brush = QtGui.QBrush(QtGui.QColor(120, 120, 120))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 255))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.BrightText, brush)
        brush = QtGui.QBrush(QtGui.QColor(120, 120, 120))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(240, 240, 240))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(240, 240, 240))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Shadow, brush)
        brush = QtGui.QBrush(QtGui.QColor(240, 240, 240))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.AlternateBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(255, 255, 220))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ToolTipBase, brush)
        brush = QtGui.QBrush(QtGui.QColor(0, 0, 0))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ToolTipText, brush)
        self.table_plugin_settings.setPalette(palette)
        font = QtGui.QFont()
        font.setFamily("Roboto")
        font.setPointSize(-1)
        font.setBold(False)
        font.setWeight(37)
        self.table_plugin_settings.setFont(font)
        self.table_plugin_settings.setStyleSheet(
            "QTableWidget { border: 1px solid #ccc; font-family: Roboto; font-size: 11px; }"
        )
        self.table_plugin_settings.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.table_plugin_settings.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.table_plugin_settings.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.table_plugin_settings.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.table_plugin_settings.setProperty("showDropIndicator", False)
        self.table_plugin_settings.setDragDropOverwriteMode(False)
        self.table_plugin_settings.setAlternatingRowColors(True)
        self.table_plugin_settings.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.table_plugin_settings.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.table_plugin_settings.setTextElideMode(QtCore.Qt.ElideNone)
        self.table_plugin_settings.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.table_plugin_settings.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.table_plugin_settings.setShowGrid(False)
        self.table_plugin_settings.setGridStyle(QtCore.Qt.NoPen)
        self.table_plugin_settings.setWordWrap(True)
        self.table_plugin_settings.setCornerButtonEnabled(True)
        self.table_plugin_settings.setObjectName("table_plugin_settings")
        self.table_plugin_settings.setColumnCount(4)
        self.table_plugin_settings.setRowCount(2)
        item = QtWidgets.QTableWidgetItem()
        item.setTextAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter | QtCore.Qt.AlignCenter)
        self.table_plugin_settings.setVerticalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setVerticalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        item.setTextAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter | QtCore.Qt.AlignCenter)
        self.table_plugin_settings.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setHorizontalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setHorizontalHeaderItem(3, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setItem(0, 0, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setItem(0, 1, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setItem(0, 2, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setItem(0, 3, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setItem(1, 1, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setItem(1, 2, item)
        item = QtWidgets.QTableWidgetItem()
        self.table_plugin_settings.setItem(1, 3, item)
        self.table_plugin_settings.horizontalHeader().setVisible(True)
        self.table_plugin_settings.horizontalHeader().setCascadingSectionResizes(False)
        self.table_plugin_settings.horizontalHeader().setHighlightSections(True)
        self.table_plugin_settings.horizontalHeader().setSortIndicatorShown(False)
        self.table_plugin_settings.horizontalHeader().setStretchLastSection(True)
        self.table_plugin_settings.verticalHeader().setVisible(False)
        self.table_plugin_settings.verticalHeader().setCascadingSectionResizes(False)
        self.table_plugin_settings.verticalHeader().setDefaultSectionSize(45)
        self.table_plugin_settings.verticalHeader().setSortIndicatorShown(False)
        self.table_plugin_settings.verticalHeader().setStretchLastSection(False)
        self.lbl_plugin_name = QtWidgets.QLabel(self.frame_details)
        self.lbl_plugin_name.setGeometry(QtCore.QRect(20, 20, 391, 21))
        font = QtGui.QFont()
        font.setFamily("Roboto")
        font.setPointSize(-1)
        font.setBold(False)
        font.setWeight(37)
        self.lbl_plugin_name.setFont(font)
        self.lbl_plugin_name.setStyleSheet("color: #555; border: 0;  font-size: 15px;")
        self.lbl_plugin_name.setTextFormat(QtCore.Qt.RichText)
        self.lbl_plugin_name.setObjectName("lbl_plugin_name")
        self.btn_reset_plugin_settings = QtWidgets.QPushButton(self.frame_details)
        self.btn_reset_plugin_settings.setGeometry(QtCore.QRect(520, 60, 180, 31))
        self.btn_reset_plugin_settings.setStyleSheet(
            "QPushButton {\n"
            "color: #eee;\n"
            "background: #666;\n"
            "border: 0; \n"
            "}\n"
            "\n"
            "QPushButton:hover {\n"
            "    background: #555;\n"
            "}\n"
            "\n"
            "QPushButton:pressed {\n"
            "    background: #111;\n"
            "}"
        )
        self.btn_reset_plugin_settings.setObjectName("btn_reset_plugin_settings")
        self.frame = QtWidgets.QFrame(self.centralwidget)
        self.frame.setGeometry(QtCore.QRect(0, 60, 281, 621))
        self.frame.setAutoFillBackground(False)
        self.frame.setStyleSheet("QFrame {background: #eee;}")
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.lbl_hint_restart = QtWidgets.QLabel(self.frame)
        self.lbl_hint_restart.setGeometry(QtCore.QRect(20, 560, 251, 51))
        palette = QtGui.QPalette()
        brush = QtGui.QBrush(QtGui.QColor(153, 153, 153))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(238, 238, 238))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(153, 153, 153))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(153, 153, 153))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(238, 238, 238))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(238, 238, 238))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Active, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(153, 153, 153))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(238, 238, 238))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(153, 153, 153))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(153, 153, 153))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(238, 238, 238))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(238, 238, 238))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Inactive, QtGui.QPalette.Window, brush)
        brush = QtGui.QBrush(QtGui.QColor(153, 153, 153))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.WindowText, brush)
        brush = QtGui.QBrush(QtGui.QColor(238, 238, 238))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Button, brush)
        brush = QtGui.QBrush(QtGui.QColor(153, 153, 153))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Text, brush)
        brush = QtGui.QBrush(QtGui.QColor(153, 153, 153))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.ButtonText, brush)
        brush = QtGui.QBrush(QtGui.QColor(238, 238, 238))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Base, brush)
        brush = QtGui.QBrush(QtGui.QColor(238, 238, 238))
        brush.setStyle(QtCore.Qt.SolidPattern)
        palette.setBrush(QtGui.QPalette.Disabled, QtGui.QPalette.Window, brush)
        self.lbl_hint_restart.setPalette(palette)
        font = QtGui.QFont()
        font.setFamily("Roboto")
        font.setBold(False)
        font.setWeight(37)
        self.lbl_hint_restart.setFont(font)
        self.lbl_hint_restart.setStyleSheet("border-top: 1px dotted #aaa; color: #999; padding: 0;")
        self.lbl_hint_restart.setTextFormat(QtCore.Qt.PlainText)
        self.lbl_hint_restart.setScaledContents(False)
        self.lbl_hint_restart.setAlignment(QtCore.Qt.AlignLeading | QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
        self.lbl_hint_restart.setWordWrap(True)
        self.lbl_hint_restart.setObjectName("lbl_hint_restart")
        self.lbl_restart_pipeline = QtWidgets.QLabel(self.centralwidget)
        self.lbl_restart_pipeline.setGeometry(QtCore.QRect(687, 18, 294, 28))
        font = QtGui.QFont()
        font.setFamily("Roboto")
        font.setPointSize(-1)
        font.setBold(False)
        font.setWeight(37)
        self.lbl_restart_pipeline.setFont(font)
        self.lbl_restart_pipeline.setAutoFillBackground(False)
        self.lbl_restart_pipeline.setStyleSheet(
            "background: #ff9719; color: #eee; padding: 5px; border: 0; font-size: 12px;"
        )
        self.lbl_restart_pipeline.setAlignment(QtCore.Qt.AlignCenter)
        self.lbl_restart_pipeline.setObjectName("lbl_restart_pipeline")
        self.label = QtWidgets.QLabel(self.centralwidget)
        self.label.setGeometry(QtCore.QRect(20, 25, 261, 31))
        self.label.setStyleSheet("font-size: 20px; color: #ff9719;\n" "font-weight: 100;")
        self.label.setObjectName("label")
        self.label_9 = QtWidgets.QLabel(self.centralwidget)
        self.label_9.setGeometry(QtCore.QRect(20, 6, 261, 31))
        self.label_9.setStyleSheet("font-size: 13px;\n" "color: #aaa;\n" "font-weight: 600;")
        self.label_9.setObjectName("label_9")
        self.frame.raise_()
        self.lst_plugins.raise_()
        self.lbl_restart_pipeline.raise_()
        self.frame_details.raise_()
        self.label.raise_()
        self.label_9.raise_()
        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Plugin Configurator"))
        __sortingEnabled = self.lst_plugins.isSortingEnabled()
        self.lst_plugins.setSortingEnabled(False)
        item = self.lst_plugins.item(0)
        item.setText(_translate("MainWindow", "Ambient Occlusion"))
        item = self.lst_plugins.item(1)
        item.setText(_translate("MainWindow", "Another Ite"))
        item = self.lst_plugins.item(2)
        item.setText(_translate("MainWindow", "Bloom"))
        item = self.lst_plugins.item(3)
        item.setText(_translate("MainWindow", "Volumetric Clouds"))
        item = self.lst_plugins.item(4)
        item.setText(_translate("MainWindow", "Depth of Field"))
        item = self.lst_plugins.item(5)
        item.setText(_translate("MainWindow", "Environment Probes"))
        item = self.lst_plugins.item(6)
        item.setText(_translate("MainWindow", "FXAA (Antialiasing)"))
        item = self.lst_plugins.item(7)
        item.setText(_translate("MainWindow", "SKAA"))
        item = self.lst_plugins.item(8)
        item.setText(_translate("MainWindow", "Motion Blur"))
        item = self.lst_plugins.item(9)
        item.setText(_translate("MainWindow", "PSSM Shadows"))
        item = self.lst_plugins.item(10)
        item.setText(_translate("MainWindow", "Atmospheric Scattering"))
        item = self.lst_plugins.item(11)
        item.setText(_translate("MainWindow", "Skin Shading"))
        item = self.lst_plugins.item(12)
        item.setText(_translate("MainWindow", "SMAA (Antialiasing)"))
        item = self.lst_plugins.item(13)
        item.setText(_translate("MainWindow", "Screen Space Reflections"))
        item = self.lst_plugins.item(14)
        item.setText(_translate("MainWindow", "Volumetric Lighting"))
        item = self.lst_plugins.item(15)
        item.setText(_translate("MainWindow", "VXGI"))
        item = self.lst_plugins.item(16)
        item.setText(_translate("MainWindow", "Item 346"))
        item = self.lst_plugins.item(17)
        item.setText(_translate("MainWindow", "Other Item"))
        item = self.lst_plugins.item(18)
        item.setText(_translate("MainWindow", "Other Item to force scrollbars"))
        self.lst_plugins.setSortingEnabled(__sortingEnabled)
        self.lbl_plugin_version.setText(_translate("MainWindow", "version 0.1 by some author"))
        self.lbl_plugin_desc.setText(
            _translate(
                "MainWindow", "This is a fancy description\n"
                "It shows information about the plugin and maybe\n"
                "a website or so."
            )
        )
        item = self.table_plugin_settings.verticalHeaderItem(0)
        item.setText(_translate("MainWindow", "Row"))
        item = self.table_plugin_settings.verticalHeaderItem(1)
        item.setText(_translate("MainWindow", "Row2"))
        item = self.table_plugin_settings.horizontalHeaderItem(0)
        item.setText(_translate("MainWindow", "Setting"))
        item = self.table_plugin_settings.horizontalHeaderItem(1)
        item.setText(_translate("MainWindow", "Default"))
        item = self.table_plugin_settings.horizontalHeaderItem(2)
        item.setText(_translate("MainWindow", "Current Value"))
        item = self.table_plugin_settings.horizontalHeaderItem(3)
        item.setText(_translate("MainWindow", "Description"))
        __sortingEnabled = self.table_plugin_settings.isSortingEnabled()
        self.table_plugin_settings.setSortingEnabled(False)
        item = self.table_plugin_settings.item(0, 0)
        item.setText(_translate("MainWindow", "Long setting name which requires"))
        item = self.table_plugin_settings.item(0, 1)
        item.setText(_translate("MainWindow", "DefaultRow0"))
        item = self.table_plugin_settings.item(0, 2)
        item.setText(_translate("MainWindow", "CurrentRow0"))
        item = self.table_plugin_settings.item(0, 3)
        item.setText(_translate("MainWindow", "DescRow0"))
        item = self.table_plugin_settings.item(1, 1)
        item.setText(_translate("MainWindow", "DefaultRow1"))
        item = self.table_plugin_settings.item(1, 2)
        item.setText(_translate("MainWindow", "CurrentRow1"))
        item = self.table_plugin_settings.item(1, 3)
        item.setText(
            _translate(
                "MainWindow",
                "aasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfgaasd\\n\\nsdf\\n\\ndfdfG\\n\\nfg"
            )
        )
        self.table_plugin_settings.setSortingEnabled(__sortingEnabled)
        self.lbl_plugin_name.setText(_translate("MainWindow", "Ambient Occlusion"))
        self.btn_reset_plugin_settings.setText(_translate("MainWindow", "Reset Settings of this Plugin"))
        self.lbl_hint_restart.setText(
            _translate("MainWindow", "Hint: Settings with a gray color require a pipeline restart when changed.")
        )
        self.lbl_restart_pipeline.setText(
            _translate("MainWindow", "Pipeline needs to be restarted to apply all changes!")
        )
        self.label.setText(_translate("MainWindow", "PLUGIN CONFIGURATOR"))
        self.label_9.setText(_translate("MainWindow", "RENDER PIPELINE"))


from . import resources_rc



"""
The existing layer names and shapes in numpy file:
(note that the terms with "value" in it are removed to save space).
default_policy/fc_1/kernel (275, 256)
default_policy/fc_1/bias (256,)
default_policy/fc_value_1/kernel (275, 256)
default_policy/fc_value_1/bias (256,)
default_policy/fc_2/kernel (256, 256)
default_policy/fc_2/bias (256,)
default_policy/fc_value_2/kernel (256, 256)
default_policy/fc_value_2/bias (256,)
default_policy/fc_out/kernel (256, 4)
default_policy/fc_out/bias (4,)
default_policy/value_out/kernel (256, 1)
default_policy/value_out/bias (1,)
"""

import os.path as osp

import numpy as np

from metadrive.engine.engine_utils import get_global_config
from metadrive.obs.state_obs import LidarStateObservation

ckpt_path = osp.join(osp.dirname(__file__), "expert_weights.npz")
_expert_weights = None
_expert_observation = None


def obs_correction(obs):
    # due to coordinate correction, this observation should be reversed
    obs[15] = 1 - obs[15]
    obs[10] = 1 - obs[10]
    return obs


def expert(vehicle, deterministic=False, need_obs=False):
    global _expert_weights
    global _expert_observation
    if _expert_weights is None:
        _expert_weights = np.load(ckpt_path)
        config = get_global_config().copy()
        config["vehicle_config"].update(
            dict(
                lidar=dict(num_lasers=240, distance=50, num_others=4, gaussian_noise=0.0, dropout_prob=0.0),
                random_agent_model=False
            )
        )
        _expert_observation = LidarStateObservation(config)
        assert _expert_observation.observation_space.shape[0] == 275, "Observation not match"
    obs = _expert_observation.observe(vehicle)
    obs = obs_correction(obs)
    weights = _expert_weights
    obs = obs.reshape(1, -1)
    x = np.matmul(obs, weights["default_policy/fc_1/kernel"]) + weights["default_policy/fc_1/bias"]
    x = np.tanh(x)
    x = np.matmul(x, weights["default_policy/fc_2/kernel"]) + weights["default_policy/fc_2/bias"]
    x = np.tanh(x)
    x = np.matmul(x, weights["default_policy/fc_out/kernel"]) + weights["default_policy/fc_out/bias"]
    x = x.reshape(-1)
    mean, log_std = np.split(x, 2)
    if deterministic:
        return (mean, obs) if need_obs else mean
    std = np.exp(log_std)
    action = np.random.normal(mean, std)
    ret = action
    # ret = np.clip(ret, -1.0, 1.0) all clip should be implemented in env!
    return (ret, obs) if need_obs else ret


def load_weights(path: str):
    """
    Load NN weights
    :param path: weights file path path
    :return: NN weights object
    """
    try:
        model = np.load(path)
        return model
    except FileNotFoundError:
        print("Can not find {}, didn't load anything".format(path))
        return None


def value(obs, weights):
    """
    Given weights, return the evaluation to one state/obseration
    :param obs: observation
    :param weights: variable weights of NN
    :return: value
    """
    if weights is None:
        return 0
    obs = obs.reshape(1, -1)
    x = np.matmul(obs, weights["default_policy/fc_value_1/kernel"]) + weights["default_policy/fc_value_1/bias"]
    x = np.tanh(x)
    x = np.matmul(x, weights["default_policy/fc_value_2/kernel"]) + weights["default_policy/fc_value_2/bias"]
    x = np.tanh(x)
    x = np.matmul(x, weights["default_policy/value_out/kernel"]) + weights["default_policy/value_out/bias"]
    ret = x.reshape(-1)
    return ret


# if __name__ == '__main__':
#     for i in range(100):
#         print("Weights? ", type(_expert_weights))
#         ret = expert(np.clip(np.random.normal(0.5, 1, size=(275,)), 0.0, 1.0))
#         print("Return: ", ret)



from collections import deque, OrderedDict

from metadrive.manager.base_manager import BaseManager
from metadrive.manager.scenario_data_manager import ScenarioDataManager


class QueueDict:
    def __init__(self, max_length):
        self.queue = deque(maxlen=max_length)
        self.dict = OrderedDict()

    def put(self, key, value):
        # Remove the key from the queue if it already exists
        if key in self.dict:
            self.queue.remove(key)
        # Else, ensure we don't exceed the max length
        elif len(self.queue) == self.queue.maxlen:
            oldest_key = self.queue.popleft()
            del self.dict[oldest_key]

        # Add the key to the queue, and update the dictionary
        self.queue.append(key)
        self.dict[key] = value

    def get(self, key):
        if key in self.dict:
            return self.dict[key]
        else:
            return None

    def items(self):
        return self.dict.items()

    def __len__(self):
        return len(self.dict)


class ScenarioCurriculumManager(BaseManager):
    PRIORITY = ScenarioDataManager.PRIORITY - 1

    def __init__(self):
        super().__init__()
        curriculum_level = self.engine.global_config["curriculum_level"]
        if curriculum_level > 1:
            assert self.engine.global_config["sequential_seed"], \
                "Sort and sequential seed is required for curriculum seed"
        if self.engine.global_config["episodes_to_evaluate_curriculum"] is None:
            self._episodes_to_eval = int(self.engine.global_config["num_scenarios"] / curriculum_level)
        else:
            self._episodes_to_eval = self.engine.global_config["episodes_to_evaluate_curriculum"]
        assert self._episodes_to_eval != 0, "episodes_to_evaluate_curriculum can not be 0"
        assert self._episodes_to_eval % self.engine.global_config["num_workers"
                                                                  ] == 0, "Can not be divisible by num_workers"
        self._episodes_to_eval = int(self._episodes_to_eval / self.engine.global_config["num_workers"])
        self.recent_route_completion = QueueDict(max_length=self._episodes_to_eval)
        self.recent_success = QueueDict(max_length=self._episodes_to_eval)
        self.target_success_rate = self.engine.global_config["target_success_rate"]

    def log_episode(self, success, route_completion):
        self.recent_route_completion.put(self.engine.data_manager.current_scenario_id, route_completion)
        self.recent_success.put(self.engine.data_manager.current_scenario_id, success)

    def before_reset(self):
        """
        It should be called before reseting all managers
        """
        if self.current_success_rate >= (self.target_success_rate-0.001) \
                and self.engine.current_level < self.engine.max_level - 1:
            self._level_up()

    def _level_up(self):
        self.engine.level_up()
        self.recent_route_completion = QueueDict(max_length=self._episodes_to_eval)
        self.recent_success = QueueDict(max_length=self._episodes_to_eval)
        self.engine.map_manager.clear_stored_maps()
        self.engine.data_manager.clear_stored_scenarios()

    @property
    def current_success_rate(self):
        return sum(self.recent_success.dict.values()) / self._episodes_to_eval

    @property
    def current_route_completion(self):
        return sum(self.recent_route_completion.dict.values()) / self._episodes_to_eval



import gymnasium as gym

from metadrive.envs.marl_envs.multi_agent_metadrive import MultiAgentMetaDrive


def _a(env, action):
    assert env.action_space.contains(action)
    obs, reward, terminated, truncated, info = env.step(action)
    assert env.observation_space.contains(obs)
    assert isinstance(info, dict)


def _step(env):
    try:
        obs, _ = env.reset()
        assert env.observation_space.contains(obs)
        for _ in range(5):
            _a(env, env.action_space.sample())
    finally:
        env.close()


def test_naive_multi_agent_metadrive():
    # env = MetaDriveEnv(config={"num_agents": 1})
    # assert isinstance(env.action_space, gym.spaces.Box)
    # _step(env)
    # env.close()

    env = MultiAgentMetaDrive(
        config={
            "map": "SSS",
            "num_agents": 4,
            "target_vehicle_configs": {"agent{}".format(i): {
                "spawn_longitude": i * 5
            }
                                       for i in range(4)}
        }
    )
    try:
        assert isinstance(env.action_space, gym.spaces.Dict)
        obs, _ = env.reset()
        assert isinstance(obs, dict)
        env.action_space.seed(0)
        for step in range(100):
            a = env.action_space.sample()
            assert isinstance(a, dict)
            o, r, tm, tc, i = env.step(a)

            pos_z_list = [v.chassis.getNode(0).transform.pos[2] for v in env.vehicles.values()]
            for p in pos_z_list:
                assert p < 5.0 or step <= 10

            assert isinstance(o, dict)
            assert isinstance(r, dict)
            assert isinstance(tm, dict)
            assert isinstance(tc, dict)
            assert isinstance(i, dict)
            if tm["__all__"]:
                break

        _step(env)
    finally:
        env.close()


if __name__ == '__main__':
    test_naive_multi_agent_metadrive()



import metadrive.register
from metadrive.envs import MetaDriveEnv, TopDownMetaDrive, TopDownSingleFrameMetaDriveEnv, TopDownMetaDriveEnvV2, \
    SafeMetaDriveEnv, MultiAgentRoundaboutEnv, MultiAgentIntersectionEnv, MultiAgentParkingLotEnv, \
    MultiAgentTollgateEnv, MultiAgentBottleneckEnv, MultiAgentMetaDrive
from metadrive.utils.registry import get_metadrive_class
import os

MetaDrive_PACKAGE_DIR = os.path.dirname(os.path.abspath(__file__))



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

# This tool offers an interface to modify materials
# Hacked together in a hour, I'm sorry.

from __future__ import print_function

import os
import sys
import time
import math
import tempfile
import colorsys
from threading import Thread
from functools import partial

os.chdir(os.path.join(os.path.dirname(os.path.realpath(__file__))))
sys.path.insert(0, os.getcwd())
sys.path.insert(0, "../../")

from rplibs.six import iteritems  # noqa
from rplibs.pyqt_imports import *  # noqa

from ui.main_window_generated import Ui_MainWindow  # noqa
from rpcore.util.network_communication import NetworkCommunication  # noqa

# Allow working with an older version of the material DB.
# ONLY for debugging the viewer.
ALLOW_OUTDATED_MATERIALS = False


class MaterialData:
    def __init__(self):
        self.name = ""
        self.shading_model = 0
        self.metallic = False
        self.roughness = 0.0
        self.specular = 0.0
        self.normal_strength = 0.0
        self.shading_model_param1 = 0.0
        self.shading_model_param2 = 0.0
        self.basecolor_r = 0.6
        self.basecolor_g = 0.6
        self.basecolor_b = 0.6


class MaterialEditor(QMainWindow, Ui_MainWindow):
    """ Interface to change the plugin settings """

    SHADING_MODELS = [
        ("Default", 0, None),
        ("Emissive", 1, None),
        ("Clearcoat", 2, None),
        ("Transparent", 3, "Transparency"),
        ("Skin", 4, None),
        ("Foliage", 5, None),
    ]

    def __init__(self):
        QMainWindow.__init__(self)
        Ui_MainWindow.__init__(self)

        self.in_update = False
        self.materials = []
        self.material = MaterialData()

        self.setupUi(self)
        self.init_shading_models()
        self.init_bindings()
        self.update_material_list()
        self.on_material_selected()

    def init_bindings(self):
        qt_connect(self.cb_shading_model, "currentIndexChanged", self.read_from_ui)
        qt_connect(self.cb_material, "currentIndexChanged", self.on_material_selected)
        qt_connect(self.cb_metallic, "stateChanged", self.read_from_ui)

        self.sliders = [
            (self.slider_roughness, self.lbl_roughness, 0.0, 1.0, "roughness"),
            (self.slider_specular, self.lbl_specular, 1.0, 2.51, "specular"),
            (self.slider_normal, self.lbl_normal, 0.0, 1.0, "normal_strength"),
            (self.slider_param1, self.lbl_param1, 0.0, 1.0, "shading_model_param1"),
        ]

        for slider, lbl, start, end, prop in self.sliders:
            qt_connect(slider, "valueChanged", self.read_from_ui)

        qt_connect(self.basecolor_1, "valueChanged", self.read_from_ui)
        qt_connect(self.basecolor_2, "valueChanged", self.read_from_ui)
        qt_connect(self.basecolor_3, "valueChanged", self.read_from_ui)

        for cb in (self.cb_rgb, self.cb_srgb, self.cb_hsv):
            qt_connect(cb, "toggled", self.write_to_ui)

    def update_ui(self):
        # Basecolor
        labels = "R", "G", "B"
        if self.cb_hsv.isChecked():
            labels = "H", "S", "V"
        self.lbl_basecolor1.setText(labels[0])
        self.lbl_basecolor2.setText(labels[1])
        self.lbl_basecolor3.setText(labels[2])

        a, b, c = (self.basecolor_1.value() / 100.0, self.basecolor_2.value() / 100.0, self.basecolor_3.value() / 100.0)
        rgb = self.tuple_to_basecolor(a, b, c)
        self.lbl_basecolor_val1.setText("{:0.2f}".format(a))
        self.lbl_basecolor_val2.setText("{:0.2f}".format(b))
        self.lbl_basecolor_val3.setText("{:0.2f}".format(c))
        self.lbl_color_preview.setStyleSheet(
            "background: rgb({}, {}, {});".format(int(rgb[0] * 255), int(rgb[1] * 255), int(rgb[2] * 255))
        )

        # Shading model
        self._update_shading_model()

    def read_from_ui(self):
        if self.in_update:
            return

        # Rest of sliders
        for slider, lbl, start, end, prop in self.sliders:
            val = (slider.value() / 100.0) * (end - start) + start
            lbl.setText("{:0.2f}".format(val))
            setattr(self.material, prop, val)

        # Basecolor
        rgb = self._get_ui_basecolor_rgb()
        self.material.basecolor_r = rgb[0]
        self.material.basecolor_g = rgb[1]
        self.material.basecolor_b = rgb[2]

        # Metallic
        self.material.metallic = self.cb_metallic.isChecked()

        # Shading model
        self.material.shading_model = self.cb_shading_model.currentIndex()

        self.update_ui()
        self.send_update()

    def write_to_ui(self):
        self.in_update = True

        # Basecolor
        values = self.basecolor_to_tuple(self.material)
        self.basecolor_1.setValue(values[0] * 100.0)
        self.basecolor_2.setValue(values[1] * 100.0)
        self.basecolor_3.setValue(values[2] * 100.0)

        # Shading model
        self.cb_shading_model.setCurrentIndex(self.material.shading_model)

        # Metallics
        self.cb_metallic.setChecked(self.material.metallic)
        self.slider_specular.setEnabled(not self.material.metallic)

        # Rest of sliders
        for slider, lbl, start, end, prop in self.sliders:
            val = getattr(self.material, prop)
            slider.setValue((val - start) / (end - start) * 100.0)

        self.in_update = False
        self.update_ui()

    def _get_ui_basecolor_rgb(self):
        """ Extracts the RGB color which is currently edited in the UI """
        a, b, c = (self.basecolor_1.value() / 100.0, self.basecolor_2.value() / 100.0, self.basecolor_3.value() / 100.0)
        return self.tuple_to_basecolor(a, b, c)

    def update_material_list(self):
        temp_path = os.path.join(tempfile.gettempdir(), "rp_materials.data")
        print("Waiting for creation of", temp_path)
        if not ALLOW_OUTDATED_MATERIALS:
            try:
                os.remove(temp_path)
            except:
                pass
        NetworkCommunication.send_async(NetworkCommunication.MATERIAL_PORT, "dump_materials " + temp_path)
        start_time = time.time()
        while not os.path.isfile(temp_path) and time.time() - start_time < 5.0:
            time.sleep(0.5)
        if not os.path.isfile(temp_path):
            QMessageBox.critical(
                self, "Error",
                "Render Pipeline not responding! Make sure a render pipeline application is running, and try again later."
            )
            sys.exit(-1)
        if not ALLOW_OUTDATED_MATERIALS:
            time.sleep(0.5)
        self._load_material_list(temp_path)

    def _load_material_list(self, pth):
        self.materials = []
        self.cb_material.clear()
        with open(pth) as handle:
            for line in handle.readlines():
                parts = line.strip().split(" ")
                material = self._read_in_material(parts)
                self.materials.append(material)
                self.cb_material.addItem(material.name)

    def _read_in_material(self, parts):
        material = MaterialData()
        material.name = parts[0]
        material.basecolor_r = float(parts[1])
        material.basecolor_g = float(parts[2])
        material.basecolor_b = float(parts[3])
        material.roughness = float(parts[4])
        material.specular = float(parts[5])
        material.metallic = float(parts[6]) > 0.5
        material.shading_model = int(float(parts[7]))
        material.normal_strength = float(parts[8])
        material.shading_model_param1 = float(parts[9])
        material.shading_model_param2 = float(parts[10])
        return material

    def basecolor_to_tuple(self, mat):
        def to_srgb(v):
            return math.pow(v, 1.0 / 2.2)

        if self.cb_rgb.isChecked():
            return mat.basecolor_r, mat.basecolor_g, mat.basecolor_b
        elif self.cb_srgb.isChecked():
            return to_srgb(mat.basecolor_r), to_srgb(mat.basecolor_g), to_srgb(mat.basecolor_b)
        elif self.cb_hsv.isChecked():
            return colorsys.rgb_to_hsv(mat.basecolor_r, mat.basecolor_g, mat.basecolor_b)
        else:
            assert False

    def tuple_to_basecolor(self, a, b, c):
        def from_srgb(v):
            return math.pow(v, 2.2)

        if self.cb_rgb.isChecked():
            return a, b, c
        elif self.cb_srgb.isChecked():
            return from_srgb(a), from_srgb(b), from_srgb(c)
        elif self.cb_hsv.isChecked():
            return colorsys.hsv_to_rgb(a, b, c)
        else:
            assert False

    def on_material_selected(self):
        index = self.cb_material.currentIndex()
        if index < 0 or index >= len(self.materials):
            print("Invalid material with index", index, "only have", len(self.materials), "materials")
            return
        self.material = self.materials[index]
        print("Loaded material", self.material.name)
        self.write_to_ui()

    def send_update(self):
        serialized = ("{} " * 11).format(
            self.material.name,
            self.material.basecolor_r,
            self.material.basecolor_g,
            self.material.basecolor_b,
            self.material.roughness,
            self.material.specular,
            1.0 if self.material.metallic else 0.0,
            self.material.shading_model,
            self.material.normal_strength,
            self.material.shading_model_param1,
            self.material.shading_model_param2,
        )
        NetworkCommunication.send_async(NetworkCommunication.MATERIAL_PORT, "update_material " + serialized)

    def _update_shading_model(self):
        name, val, optional_param = self.SHADING_MODELS[self.cb_shading_model.currentIndex()]
        if optional_param is None:
            self.slider_param1.setEnabled(False)
            self.lbl_shading_model_param1.setText("Unused")
        else:
            self.slider_param1.setEnabled(True)
            self.lbl_shading_model_param1.setText(optional_param)

        self.cb_metallic.show()
        self.slider_roughness.setEnabled(True)
        self.slider_normal.setEnabled(True)

        if name == "Emissive":
            self.cb_metallic.hide()
            self.slider_roughness.setEnabled(False)
            self.slider_specular.setEnabled(False)
            self.slider_normal.setEnabled(False)
        elif name == "Clearcoat":
            self.cb_metallic.hide()
            self.slider_specular.setEnabled(False)
        elif name == "Transparent":
            pass
        elif name == "Skin":
            self.cb_metallic.hide()
        elif name == "Foliage":
            self.cb_metallic.hide()

    def init_shading_models(self):
        self.cb_shading_model.clear()
        for name, val, optional_param in self.SHADING_MODELS:
            self.cb_shading_model.addItem(name)


# Start application
app = QApplication(sys.argv)
qt_register_fonts()
editor = MaterialEditor()
editor.show()
app.exec_()



import math
import numpy as np

from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.pgblock.create_pg_block_utils import ExtendStraightLane, CreateRoadFrom, CreateAdverseRoad, \
    create_bend_straight
from metadrive.component.pgblock.pg_block import PGBlock
from metadrive.component.road_network import Road
from metadrive.constants import Decoration, PGLineType
from metadrive.utils.pg.utils import check_lane_on_road
from metadrive.component.pg_space import ParameterSpace, Parameter, BlockParameterSpace


class Ramp(PGBlock):
    """
                    InRamp                                             OutRamp
    
     start ----------- end ------------------           start ----------- end ------------------
     start ----------- end ------------------           start ----------- end ------------------              
    (start ----------- end)[----------------]           start ----------- end [----------------]
                       end -----------------}          (start ---------- {end) 
                      //                                                      \\                                                         
    { ---------------//                                                        \\---------------}    
    """
    PARAMETER_SPACE = ParameterSpace(BlockParameterSpace.RAMP_PARAMETER)
    SOCKET_NUM = 1

    # can be added to parameter space in the future
    RADIUS = 40  # meters
    ANGLE = 10  # degree
    LANE_TYPE = (PGLineType.CONTINUOUS, PGLineType.CONTINUOUS)
    SPEED_LIMIT = 12  # 12 m/s ~= 40 km/h
    CONNECT_PART_LEN = 20
    RAMP_LEN = 15


class InRampOnStraight(Ramp):
    ID = "r"
    EXTRA_PART = 10
    SOCKET_LEN = 20

    def _try_plug_into_previous_block(self) -> bool:
        acc_lane_len = self.get_config()[Parameter.length]
        no_cross = True

        # extend road and acc raod part, part 0
        self.set_part_idx(0)
        sin_angle = math.sin(np.deg2rad(self.ANGLE))
        cos_angle = math.cos(np.deg2rad(self.ANGLE))
        longitude_len = sin_angle * self.RADIUS * 2 + cos_angle * self.CONNECT_PART_LEN + self.RAMP_LEN

        extend_lane = ExtendStraightLane(
            self.positive_basic_lane, longitude_len + self.EXTRA_PART, [PGLineType.BROKEN, PGLineType.CONTINUOUS]
        )
        extend_road = Road(self.pre_block_socket.positive_road.end_node, self.add_road_node())
        no_cross = CreateRoadFrom(
            extend_lane,
            self.positive_lane_num,
            extend_road,
            self.block_network,
            self._global_network,
            side_lane_line_type=PGLineType.CONTINUOUS,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        extend_road.get_lanes(self.block_network)[-1].line_types = [
            PGLineType.BROKEN if self.positive_lane_num != 1 else PGLineType.CONTINUOUS, PGLineType.CONTINUOUS
        ]
        no_cross = CreateAdverseRoad(
            extend_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        _extend_road = -extend_road
        left_lane_line = PGLineType.NONE if self.positive_lane_num == 1 else PGLineType.BROKEN
        _extend_road.get_lanes(self.block_network)[-1].line_types = [left_lane_line, PGLineType.SIDE]

        # main acc part
        acc_side_lane = ExtendStraightLane(
            extend_lane, acc_lane_len + self.lane_width, [extend_lane.line_types[0], PGLineType.SIDE]
        )
        acc_road = Road(extend_road.end_node, self.add_road_node())
        no_cross = CreateRoadFrom(
            acc_side_lane,
            self.positive_lane_num,
            acc_road,
            self.block_network,
            self._global_network,
            side_lane_line_type=PGLineType.CONTINUOUS,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        no_cross = CreateAdverseRoad(
            acc_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        left_line_type = PGLineType.CONTINUOUS if self.positive_lane_num == 1 else PGLineType.BROKEN
        acc_road.get_lanes(self.block_network)[-1].line_types = [left_line_type, PGLineType.BROKEN]

        # socket part
        socket_side_lane = ExtendStraightLane(acc_side_lane, self.SOCKET_LEN, acc_side_lane.line_types)
        socket_road = Road(acc_road.end_node, self.add_road_node())
        no_cross = CreateRoadFrom(
            socket_side_lane,
            self.positive_lane_num,
            socket_road,
            self.block_network,
            self._global_network,
            side_lane_line_type=PGLineType.CONTINUOUS,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        no_cross = CreateAdverseRoad(
            socket_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        self.add_sockets(PGBlock.create_socket_from_positive_road(socket_road))

        # ramp part, part 1
        self.set_part_idx(1)
        lateral_dist = (1 - cos_angle) * self.RADIUS * 2 + sin_angle * self.CONNECT_PART_LEN
        end_point = extend_lane.position(self.EXTRA_PART + self.RAMP_LEN, lateral_dist + self.lane_width)
        start_point = extend_lane.position(self.EXTRA_PART, lateral_dist + self.lane_width)
        straight_part = StraightLane(
            start_point, end_point, self.lane_width, self.LANE_TYPE, speed_limit=self.SPEED_LIMIT
        )
        straight_road = Road(self.add_road_node(), self.add_road_node())
        self.block_network.add_lane(straight_road.start_node, straight_road.end_node, straight_part)
        no_cross = (
            not check_lane_on_road(
                self._global_network,
                straight_part,
                0.95,
                ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and no_cross
        self.add_respawn_roads(straight_road)

        # p1 road 0, 1
        bend_1, connect_part = create_bend_straight(
            straight_part,
            self.CONNECT_PART_LEN,
            self.RADIUS,
            np.deg2rad(self.ANGLE),
            False,
            self.lane_width,
            self.LANE_TYPE,
            speed_limit=self.SPEED_LIMIT
        )
        bend_1_road = Road(straight_road.end_node, self.add_road_node())
        connect_road = Road(bend_1_road.end_node, self.add_road_node())
        self.block_network.add_lane(bend_1_road.start_node, bend_1_road.end_node, bend_1)
        self.block_network.add_lane(connect_road.start_node, connect_road.end_node, connect_part)
        no_cross = (
            not check_lane_on_road(
                self._global_network, bend_1, 0.95, ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and no_cross
        no_cross = (
            not check_lane_on_road(
                self._global_network,
                connect_part,
                0.95,
                ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and no_cross

        # p1, road 2, 3
        bend_2, acc_lane = create_bend_straight(
            connect_part,
            acc_lane_len,
            self.RADIUS,
            np.deg2rad(self.ANGLE),
            True,
            self.lane_width,
            self.LANE_TYPE,
            speed_limit=self.SPEED_LIMIT
        )
        acc_lane.line_types = [PGLineType.BROKEN, PGLineType.CONTINUOUS]
        bend_2_road = Road(connect_road.end_node, self.road_node(0, 0))  # end at part1 road 0, extend road
        self.block_network.add_lane(bend_2_road.start_node, bend_2_road.end_node, bend_2)
        self.block_network.add_lane(acc_road.start_node, acc_road.end_node, acc_lane)
        no_cross = (
            not check_lane_on_road(
                self._global_network, bend_2, 0.95, ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and no_cross
        no_cross = (
            not check_lane_on_road(
                self._global_network, acc_lane, 0.95, ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and no_cross

        # p1, road 4, small circular to decorate
        merge_lane, _ = create_bend_straight(
            acc_lane, 10, self.lane_width / 2, np.pi / 2, False, self.lane_width,
            (PGLineType.BROKEN, PGLineType.CONTINUOUS)
        )
        self.block_network.add_lane(Decoration.start, Decoration.end, merge_lane)

        return no_cross

    def get_intermediate_spawn_lanes(self):
        """
        Remove lanes on socket
        """
        spawn_lanes = super(InRampOnStraight, self).get_intermediate_spawn_lanes()
        lane_on_socket = self.get_socket(0).get_positive_lanes(self.block_network)[0]
        filtered = []
        for lanes in spawn_lanes:
            if lane_on_socket in lanes:
                continue
            else:
                filtered.append(lanes)
        spawn_lanes = filtered
        assert sum([abs(l.length - InRampOnStraight.RAMP_LEN) <= 0.1 for ls in spawn_lanes for l in ls]) == 1
        return spawn_lanes


class OutRampOnStraight(Ramp):
    ID = "R"
    EXTRA_LEN = 15

    def _get_deacc_lane(self, att_lane: StraightLane):
        start = att_lane.position(self.lane_width, self.lane_width)
        end = att_lane.position(att_lane.length, self.lane_width)
        return StraightLane(start, end, self.lane_width, (PGLineType.BROKEN, PGLineType.CONTINUOUS))

    def _get_merge_part(self, side_lane: StraightLane):
        tool_lane = StraightLane(side_lane.end, side_lane.start, side_lane.width)
        merge_part, _ = create_bend_straight(
            tool_lane,
            10,
            self.lane_width / 2,
            np.pi / 2,
            True,
            width=self.lane_width,
            line_types=(PGLineType.CONTINUOUS, PGLineType.BROKEN)
        )
        return merge_part

    def _try_plug_into_previous_block(self) -> bool:
        no_cross = True
        sin_angle = math.sin(np.deg2rad(self.ANGLE))
        cos_angle = math.cos(np.deg2rad(self.ANGLE))
        longitude_len = sin_angle * self.RADIUS * 2 + cos_angle * self.CONNECT_PART_LEN + self.RAMP_LEN + self.EXTRA_LEN

        self.set_part_idx(0)
        # part 0 road 0
        dec_lane_len = self.get_config()[Parameter.length]
        dec_lane = ExtendStraightLane(
            self.positive_basic_lane, dec_lane_len + self.lane_width,
            [self.positive_basic_lane.line_types[0], PGLineType.SIDE]
        )
        dec_road = Road(self.pre_block_socket.positive_road.end_node, self.add_road_node())
        no_cross = CreateRoadFrom(
            dec_lane,
            self.positive_lane_num,
            dec_road,
            self.block_network,
            self._global_network,
            side_lane_line_type=PGLineType.CONTINUOUS,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        no_cross = CreateAdverseRoad(
            dec_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        dec_right_lane = dec_road.get_lanes(self.block_network)[-1]
        left_line_type = PGLineType.CONTINUOUS if self.positive_lane_num == 1 else PGLineType.BROKEN
        dec_right_lane.line_types = [left_line_type, PGLineType.BROKEN]

        # part 0 road 1
        extend_lane = ExtendStraightLane(
            dec_right_lane, longitude_len, [dec_right_lane.line_types[0], PGLineType.CONTINUOUS]
        )
        extend_road = Road(dec_road.end_node, self.add_road_node())
        no_cross = CreateRoadFrom(
            extend_lane,
            self.positive_lane_num,
            extend_road,
            self.block_network,
            self._global_network,
            side_lane_line_type=PGLineType.CONTINUOUS,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        no_cross = CreateAdverseRoad(
            extend_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        _extend_road = -extend_road
        _extend_road.get_lanes(self.block_network)[-1].line_types = [
            PGLineType.NONE if self.positive_lane_num == 1 else PGLineType.BROKEN, PGLineType.SIDE
        ]
        self.add_sockets(self.create_socket_from_positive_road(extend_road))

        # part 1 road 0
        self.set_part_idx(1)
        dec_side_right_lane = self._get_deacc_lane(dec_right_lane)
        self.block_network.add_lane(dec_road.start_node, dec_road.end_node, dec_side_right_lane)
        no_cross = (
            not check_lane_on_road(
                self._global_network,
                dec_side_right_lane,
                0.95,
                ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and no_cross

        bend_1, connect_part = create_bend_straight(
            dec_side_right_lane,
            self.CONNECT_PART_LEN,
            self.RADIUS,
            np.deg2rad(self.ANGLE),
            True,
            self.lane_width,
            self.LANE_TYPE,
            speed_limit=self.SPEED_LIMIT
        )
        bend_1_road = Road(dec_road.end_node, self.add_road_node())
        connect_road = Road(bend_1_road.end_node, self.add_road_node())
        self.block_network.add_lane(bend_1_road.start_node, bend_1_road.end_node, bend_1)
        self.block_network.add_lane(connect_road.start_node, connect_road.end_node, connect_part)
        no_cross = (
            not check_lane_on_road(
                self._global_network, bend_1, 0.95, ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and no_cross
        no_cross = (
            not check_lane_on_road(
                self._global_network,
                connect_part,
                0.95,
                ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and no_cross

        bend_2, straight_part = create_bend_straight(
            connect_part,
            self.RAMP_LEN,
            self.RADIUS,
            np.deg2rad(self.ANGLE),
            False,
            self.lane_width,
            self.LANE_TYPE,
            speed_limit=self.SPEED_LIMIT
        )
        bend_2_road = Road(connect_road.end_node, self.add_road_node())  # end at part1 road 0, extend road
        straight_road = Road(bend_2_road.end_node, self.add_road_node())
        self.block_network.add_lane(bend_2_road.start_node, bend_2_road.end_node, bend_2)
        self.block_network.add_lane(straight_road.start_node, straight_road.end_node, straight_part)
        no_cross = (
            not check_lane_on_road(
                self._global_network, bend_2, 0.95, ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and no_cross
        no_cross = (
            not check_lane_on_road(
                self._global_network,
                straight_part,
                0.95,
                ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and no_cross

        decoration_part = self._get_merge_part(dec_side_right_lane)
        self.block_network.add_lane(Decoration.start, Decoration.end, decoration_part)
        return no_cross



import inspect
import os
from typing import Callable
from metadrive.type import MetaDriveType
from metadrive.engine.asset_loader import AssetLoader
from metadrive.scenario.scenario_description import ScenarioDescription as SD
from metadrive.scenario.utils import read_scenario_data, read_dataset_summary


def test_read_data_and_create_summary():
    # waymo = AssetLoader.file_path("waymo", return_raw_style=False)
    nuscenes = AssetLoader.file_path("nuscenes", return_raw_style=False)
    summary, scenarios, mapping = read_dataset_summary(nuscenes)
    sd_scenario = read_scenario_data(os.path.join(nuscenes, mapping[scenarios[2]], scenarios[2]))

    summary_dict = {}
    for track_id, track in sd_scenario[SD.TRACKS].items():
        summary_dict[track_id] = SD.get_object_summary(state_dict=track, id=track_id)
    sd_scenario[SD.METADATA][SD.SUMMARY.OBJECT_SUMMARY] = summary_dict

    # test
    assert sd_scenario[SD.METADATA][SD.SDC_ID] in summary_dict

    # count some objects occurrence
    sd_scenario[SD.METADATA][SD.SUMMARY.NUMBER_SUMMARY] = SD.get_number_summary(sd_scenario)

    all_v = sd_scenario[SD.METADATA][SD.SUMMARY.NUMBER_SUMMARY][SD.SUMMARY.NUM_OBJECTS_EACH_TYPE][MetaDriveType.VEHICLE]
    moving_v = sd_scenario[SD.METADATA][SD.SUMMARY.NUMBER_SUMMARY][SD.SUMMARY.NUM_MOVING_OBJECTS_EACH_TYPE][
        MetaDriveType.VEHICLE]
    static_v = all_v - moving_v
    assert static_v == 14, "Wrong calculation"


def test_repeated_key_check():
    def _check_dict_variable(d):
        attributes = {key: value for key, value in d.__dict__.items() if not key.startswith('__') and not callable(key)}
        keys = set()
        values = set()
        for k, v in attributes.items():
            if isinstance(v, str):
                keys.add(k)
                values.add(v)
        assert len(keys) == len(values)

    _check_dict_variable(SD)
    _check_dict_variable(SD.SUMMARY)


if __name__ == '__main__':
    # test_repeated_key_check()
    test_read_data_and_create_summary()



from panda3d.core import loadPrcFileData
from metadrive.tests.test_installation import verify_installation
import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--cuda", action="store_true")
    parser.add_argument("--camera", type=str, default="main", choices=["main", "rgb", "depth"])
    args = parser.parse_args()
    loadPrcFileData("", "notify-level-task fatal")
    verify_installation(args.cuda, args.camera)



import pickle

from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger


def vis_gen_map_read():
    """
    For visualization check
    """
    env_num = 3
    step = 500
    generate_config = {"num_scenarios": env_num, "start_seed": 0, "use_render": True}
    restore_config = {"num_scenarios": env_num, "start_seed": 0, "use_render": True}

    setup_logger(debug=True)
    try:
        env = MetaDriveEnv(generate_config)
        env.reset()
        data = env.engine.map_manager.dump_all_maps(file_name="test_10maps.pickle")
        for i in range(step):
            env.step(env.action_space.sample())
        env.close()

        # Check load
        with open("test_10maps.pickle", "rb") as f:
            restored_data = pickle.load(f)

        env = MetaDriveEnv(restore_config)
        env.reset()
        env.engine.map_manager.load_all_maps("test_10maps.pickle")
        env.reset()

        for i in range(step):
            env.step(env.action_space.sample())
    finally:
        env.close()


if __name__ == "__main__":
    vis_gen_map_read()



import numpy as np
from shapely import geometry

from metadrive.component.lane.abs_lane import AbstractLane
from metadrive.constants import DrivableAreaProperty


class PGLane(AbstractLane):
    POLYGON_SAMPLE_RATE = 1
    radius = 0.0

    def __init__(self):
        super(PGLane, self).__init__()
        # one should implement how to get polygon in property def polygon(self)
        self._polygon = None
        self._shapely_polygon = None

    @property
    def shapely_polygon(self):
        if self._shapely_polygon is None:
            assert self._polygon is not None
            self._shapely_polygon = geometry.Polygon(geometry.LineString(self._polygon))
        return self._shapely_polygon

    def construct_sidewalk(self, block, lateral):
        if block.use_render_pipeline:
            # donot construct sidewalk
            return
        radius = self.radius
        segment_num = int(self.length / DrivableAreaProperty.SIDEWALK_LENGTH)
        for segment in range(segment_num):
            lane_start = self.position(segment * DrivableAreaProperty.SIDEWALK_LENGTH, lateral)
            lane_end = self.position((segment + 1) * DrivableAreaProperty.SIDEWALK_LENGTH, lateral)
            if segment == segment_num - 1:
                lane_end = self.position(self.length, lateral)
            if radius == 0:
                factor = 1
            else:
                if self.is_clockwise():
                    factor = (1 - block.SIDEWALK_LINE_DIST / radius)
                else:
                    factor = (1 + block.SIDEWALK_WIDTH / radius) * (1 + block.SIDEWALK_LINE_DIST / radius)
            node_path_list = self.construct_sidewalk_segment(
                block,
                lane_start,
                lane_end,
                length_multiply=factor,
                extra_thrust=DrivableAreaProperty.SIDEWALK_WIDTH / 2 + DrivableAreaProperty.SIDEWALK_LINE_DIST
            )
            self._node_path_list.extend(node_path_list)

    @property
    def polygon(self):
        raise NotImplementedError("Overwrite this function to allow getting polygon for this lane")

    def point_on_lane(self, point):
        s_point = geometry.Point(point[0], point[1])
        return self.shapely_polygon.contains(s_point)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import math

from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error
from metadrive.render_pipeline.rplibs.six import iteritems, itervalues

from direct.stdpy.file import listdir, isfile, join
from panda3d.core import SamplerState, ShaderAttrib, NodePath

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.loader import RPLoader


class ScatteringMethod(RPObject):
    """ Base class for all scattering methods """
    def __init__(self, plugin_handle):
        RPObject.__init__(self)
        self.handle = plugin_handle

    def load(self):
        """ Should load all required resources here """
        raise NotImplementedError()

    def compute(self):
        """ Should compute the model here if necessary """
        raise NotImplementedError()


class ScatteringMethodHosekWilkie(ScatteringMethod):
    """ Scattering as suggested by Hosek and Wilkie """
    def load(self):
        """ Loads the scattering method """
        lut_src = self.handle.get_resource("hosek_wilkie_scattering/scattering_lut.txo")

        if not isfile(lut_src):
            self.error(
                "Could not find precompiled LUT for the Hosek Wilkie "
                "Scattering! Make sure you compiled the algorithm code!"
            )
            return

        lut_tex = RPLoader.load_sliced_3d_texture(lut_src, 512, 128, 100)
        lut_tex.set_wrap_u(SamplerState.WM_repeat)
        lut_tex.set_wrap_v(SamplerState.WM_clamp)
        lut_tex.set_wrap_w(SamplerState.WM_clamp)
        lut_tex.set_minfilter(SamplerState.FT_linear)
        lut_tex.set_magfilter(SamplerState.FT_linear)

        # Setting the format explicitely shouldn't be necessary
        # lut_tex.set_format(Image.F_rgb16)

        self.handle.display_stage.set_shader_input("ScatteringLUT", lut_tex)
        self.handle.envmap_stage.set_shader_input("ScatteringLUT", lut_tex)

    def compute(self):
        """ Computes the scattering method, not required since we use a precomputed
        LUT """
        pass


class ScatteringMethodEricBruneton(ScatteringMethod):
    """ Precomputed atmospheric scattering by Eric Bruneton """
    def load(self):
        """ Inits parameters, those should match with the ones specified in common.glsl """
        self.use_32_bit = False
        self.trans_w, self.trans_h = 256 * 4, 64 * 4
        self.sky_w, self.sky_h = 64 * 4, 16 * 4
        self.res_r, self.res_mu, self.res_mu_s, self.res_nu = 32, 128, 32, 8
        self.res_mu_s_nu = self.res_mu_s * self.res_nu

        self.create_shaders()
        self.create_textures()

    def create_textures(self):
        """ Creates all textures required for the scattering """

        tex_format = "RGBA32" if self.use_32_bit else "RGBA16"
        img_2d, img_3d = Image.create_2d, Image.create_3d

        self.textures = {
            "transmittance": img_2d("scat-trans", self.trans_w, self.trans_h, tex_format),
            "irradiance": img_2d("scat-irrad", self.sky_w, self.sky_h, tex_format),
            "inscatter": img_3d("scat-inscat", self.res_mu_s_nu, self.res_mu, self.res_r, tex_format),  # noqa # pylint: disable=line-too-long
            "delta_e": img_2d("scat-dx-e", self.sky_w, self.sky_h, tex_format),
            "delta_sr": img_3d("scat-dx-sr", self.res_mu_s_nu, self.res_mu, self.res_r, tex_format),
            "delta_sm": img_3d("scat-dx-sm", self.res_mu_s_nu, self.res_mu, self.res_r, tex_format),
            "delta_j": img_3d("scat-dx-j", self.res_mu_s_nu, self.res_mu, self.res_r, tex_format),
        }

        for img in itervalues(self.textures):
            img.set_minfilter(SamplerState.FT_linear)
            img.set_magfilter(SamplerState.FT_linear)
            img.set_wrap_u(SamplerState.WM_clamp)
            img.set_wrap_v(SamplerState.WM_clamp)
            img.set_wrap_w(SamplerState.WM_clamp)

    def create_shaders(self):
        """ Creates all the shaders used for precomputing """
        self.shaders = {}
        resource_path = self.handle.get_shader_resource("eric_bruneton")
        for fname in listdir(resource_path):
            fpath = join(resource_path, fname)
            if isfile(fpath) and fname.endswith(".compute.glsl"):
                shader_name = fname.split(".")[0]
                shader_obj = RPLoader.load_shader(fpath)
                self.shaders[shader_name] = shader_obj

    def exec_compute_shader(self, shader_obj, shader_inputs, exec_size, workgroup_size=(16, 16, 1)):
        """ Executes a compute shader. The shader object should be a shader
        loaded with Shader.load_compute, the shader inputs should be a dict where
        the keys are the names of the shader inputs and the values are the
        inputs. The workgroup_size has to match the size defined in the
        compute shader """
        ntx = int(math.ceil(exec_size[0] / workgroup_size[0]))
        nty = int(math.ceil(exec_size[1] / workgroup_size[1]))
        ntz = int(math.ceil(exec_size[2] / workgroup_size[2]))

        nodepath = NodePath("shader")
        nodepath.set_shader(shader_obj)
        nodepath.set_shader_inputs(**shader_inputs)

        attr = nodepath.get_attrib(ShaderAttrib)
        Globals.base.graphicsEngine.dispatch_compute((ntx, nty, ntz), attr, Globals.base.win.gsg)

    def compute(self):
        """ Precomputes the scattering """

        self.debug("Precomputing ...")
        exec_cshader = self.exec_compute_shader

        # Transmittance
        exec_cshader(
            self.shaders["transmittance"], {"dest": self.textures["transmittance"]}, (self.trans_w, self.trans_h, 1)
        )

        # Delta E
        exec_cshader(
            self.shaders["delta_e"], {
                "transmittanceSampler": self.textures["transmittance"],
                "dest": self.textures["delta_e"]
            }, (self.sky_w, self.sky_h, 1)
        )

        # Delta S
        exec_cshader(
            self.shaders["delta_sm_sr"], {
                "transmittanceSampler": self.textures["transmittance"],
                "destDeltaSR": self.textures["delta_sr"],
                "destDeltaSM": self.textures["delta_sm"]
            }, (self.res_mu_s_nu, self.res_mu, self.res_r), (8, 8, 8)
        )

        # Copy deltaE to irradiance texture
        exec_cshader(
            self.shaders["copy_irradiance"], {
                "k": 0.0,
                "deltaESampler": self.textures["delta_e"],
                "dest": self.textures["irradiance"]
            }, (self.sky_w, self.sky_h, 1)
        )

        # Copy delta s into inscatter texture
        exec_cshader(
            self.shaders["copy_inscatter"], {
                "deltaSRSampler": self.textures["delta_sr"],
                "deltaSMSampler": self.textures["delta_sm"],
                "dest": self.textures["inscatter"]
            }, (self.res_mu_s_nu, self.res_mu, self.res_r), (8, 8, 8)
        )

        for order in range(2, 5):
            first = order == 2

            # Delta J
            exec_cshader(
                self.shaders["delta_j"], {
                    "transmittanceSampler": self.textures["transmittance"],
                    "deltaSRSampler": self.textures["delta_sr"],
                    "deltaSMSampler": self.textures["delta_sm"],
                    "deltaESampler": self.textures["delta_e"],
                    "dest": self.textures["delta_j"],
                    "first": first
                }, (self.res_mu_s_nu, self.res_mu, self.res_r), (8, 8, 8)
            )

            # Delta E
            exec_cshader(
                self.shaders["irradiance_n"], {
                    "transmittanceSampler": self.textures["transmittance"],
                    "deltaSRSampler": self.textures["delta_sr"],
                    "deltaSMSampler": self.textures["delta_sm"],
                    "dest": self.textures["delta_e"],
                    "first": first
                }, (self.sky_w, self.sky_h, 1)
            )

            # Delta Sr
            exec_cshader(
                self.shaders["delta_sr"], {
                    "transmittanceSampler": self.textures["transmittance"],
                    "deltaJSampler": self.textures["delta_j"],
                    "dest": self.textures["delta_sr"],
                    "first": first
                }, (self.res_mu_s_nu, self.res_mu, self.res_r), (8, 8, 8)
            )

            # Add delta E to irradiance
            exec_cshader(
                self.shaders["add_delta_e"], {
                    "deltaESampler": self.textures["delta_e"],
                    "dest": self.textures["irradiance"],
                }, (self.sky_w, self.sky_h, 1)
            )

            # Add deltaSr to inscatter texture
            exec_cshader(
                self.shaders["add_delta_sr"], {
                    "deltaSSampler": self.textures["delta_sr"],
                    "dest": self.textures["inscatter"]
                }, (self.res_mu_s_nu, self.res_mu, self.res_r), (8, 8, 8)
            )

        # Make stages available
        for stage in [self.handle.display_stage, self.handle.envmap_stage]:
            stage.set_shader_inputs(
                InscatterSampler=self.textures["inscatter"],
                transmittanceSampler=self.textures["transmittance"],
                IrradianceSampler=self.textures["irradiance"]
            )



import copy
import logging
import math
from typing import Dict

import numpy as np
import seaborn as sns
from panda3d.bullet import BulletWorld, BulletBodyNode
from panda3d.core import LVector3, NodePath, PandaNode

from metadrive.base_class.base_runnable import BaseRunnable
from metadrive.constants import ObjectState
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.core.physics_world import PhysicsWorld
from metadrive.engine.physics_node import BaseRigidBodyNode, BaseGhostBodyNode
from metadrive.utils import Vector
from metadrive.utils import get_np_random
from metadrive.utils import random_string
from metadrive.utils.coordinates_shift import panda_vector, metadrive_vector, panda_heading
from metadrive.utils.math import clip
from metadrive.utils.math import norm
from metadrive.utils.math import wrap_to_pi

logger = logging.getLogger(__name__)


def _clean_a_node_path(node_path):
    if not node_path.isEmpty():
        for sub_node_path in node_path.getChildren():
            _clean_a_node_path(sub_node_path)
    node_path.detachNode()
    node_path.removeNode()


def clear_node_list(node_path_list):
    from metadrive.engine.engine_utils import get_engine
    engine = get_engine()
    for node_path in node_path_list:
        if isinstance(node_path, NodePath):
            _clean_a_node_path(node_path)
            continue

        elif isinstance(node_path, BaseRigidBodyNode):
            # PZH: Note that this line is extremely important!!!
            # It breaks the cycle reference thus we can release nodes!!!
            # It saves Waymo env!!!
            node_path.destroy()

        elif isinstance(node_path, BaseGhostBodyNode):
            # PZH: Note that this line is extremely important!!!
            # It breaks the cycle reference thus we can release nodes!!!
            # It saves Waymo env!!!
            node_path.destroy()

        elif isinstance(node_path, BulletBodyNode):
            pass

        elif isinstance(node_path, PandaNode):
            node_path.removeAllChildren()
            node_path.clearPythonTag(node_path.getName())

        else:
            raise ValueError(node_path)

        if engine is not None:
            engine.physics_world.static_world.remove(node_path)
            engine.physics_world.dynamic_world.remove(node_path)


class PhysicsNodeList(list):
    def __init__(self):
        super(PhysicsNodeList, self).__init__()
        self.attached = False

    def attach_to_physics_world(self, bullet_world: BulletWorld):
        """
        Attach the nodes in this list to bullet world
        :param bullet_world: BulletWorld()
        :return: None
        """
        if self.attached:
            return
        for node in self:
            bullet_world.attach(node)
        self.attached = True

    def detach_from_physics_world(self, bullet_world: BulletWorld):
        """
         Detach the nodes in this list from bullet world
         :param bullet_world: BulletWorld()
         :return: None
         """
        if not self.attached:
            return
        for node in self:
            bullet_world.remove(node)
        self.attached = False

    def destroy_node_list(self, bullet_world: BulletWorld):
        for node in self:
            bullet_world.remove(node)
        self.clear()


class BaseObject(BaseRunnable):
    """
    BaseObject is something interacting with game engine. If something is expected to have a body in the world or have
    appearance in the world, it must be a subclass of BaseObject.

    It is created with name/config/randomEngine and can make decision in the world. Besides the random engine can help
    sample some special configs for it ,Properties and parameters in PARAMETER_SPACE of the object are fixed after
    calling __init__().
    """
    MASS = None  # if object has a body, the mass will be set automatically
    COLLISION_MASK = None

    def __init__(self, name=None, random_seed=None, config=None, escape_random_seed_assertion=False):
        """
        Config is a static conception, which specified the parameters of one element.
        There parameters doesn't change, such as length of straight road, max speed of one vehicle, etc.
        """
        config = copy.deepcopy(config)
        super(BaseObject, self).__init__(name, random_seed, config)
        if not escape_random_seed_assertion:
            assert random_seed is not None, "Please assign a random seed for {} class.".format(self.class_name)

        # Following properties are available when this object needs visualization and physics property
        self._body = None

        # each element has its node_path to render, physics node are child nodes of it
        self.origin = NodePath(self.name)

        # Temporally store bullet nodes that have to place in bullet world (not NodePath)
        self.dynamic_nodes = PhysicsNodeList()

        # Nodes in this tuple didn't interact with other nodes! they only used to do rayTest or sweepTest
        self.static_nodes = PhysicsNodeList()

        # render or not
        self.render = False if AssetLoader.loader is None else True
        if self.render:
            self.loader = AssetLoader.get_loader()

            if not hasattr(self.loader, "loader"):
                # It is closed before!
                self.loader.__init__()

        # add color setting for visualization
        color = sns.color_palette("colorblind")
        color.remove(color[2])  # Remove the green and leave it for special vehicle
        idx = get_np_random().randint(len(color))
        rand_c = color[idx]
        self._panda_color = rand_c

        # store all NodePath reparented to this node
        self._node_path_list = []

        # debug
        self.coordinates_debug_np = None
        self.need_show_coordinates = False

    def disable_gravity(self):
        self._body.setGravity(LVector3(0, 0, 0))

    @property
    def height(self):
        return self.origin.getPos()[-1]

    @property
    def panda_color(self):
        return self._panda_color

    def add_body(self, physics_body, add_to_static_world=False):
        if self._body is None:
            # add it to physics world, in which this object will interact with other object (like collision)
            if not isinstance(physics_body, BulletBodyNode):
                raise ValueError("The physics body is not BulletBodyNode type")
            self._body = physics_body
            new_origin = NodePath(self._body)
            new_origin.setH(self.origin.getH())
            new_origin.setPos(self.origin.getPos())
            self.origin.getChildren().reparentTo(new_origin)

            # TODO(PZH): We don't call this sentence:. It might cause problem since we don't remove old origin?
            # self.origin.removeNode()

            if self.COLLISION_MASK is not None:
                self._body.setIntoCollideMask(self.COLLISION_MASK)

            self._node_path_list.append(self.origin)
            self.origin = new_origin
            if add_to_static_world:
                self.static_nodes.append(physics_body)
            else:
                self.dynamic_nodes.append(physics_body)
            if self.MASS is not None:
                assert isinstance(self.MASS,
                                  int) or isinstance(self.MASS, float), "MASS should be a float or an integer"
                self._body.setMass(self.MASS)

            if self.engine is not None and self.engine.global_config["show_coordinates"]:
                self.need_show_coordinates = True
                self.show_coordinates()
        else:
            raise AttributeError("You can not set the object body for twice")

    @property
    def body(self):
        if self._body.hasPythonTag(self._body.getName()):
            return self._body.getPythonTag(self._body.getName())
        else:
            return self._body

    def attach_to_world(self, parent_node_path: NodePath, physics_world: PhysicsWorld):
        """
        Load to world from memory
        """
        if self.render:
            # double check :-)
            assert isinstance(self.origin, NodePath), "No render model on node_path in this Element"
            self.origin.reparentTo(parent_node_path)
        self.dynamic_nodes.attach_to_physics_world(physics_world.dynamic_world)
        self.static_nodes.attach_to_physics_world(physics_world.static_world)
        logger.debug("{} is attached to the world.".format(type(self)))

    def detach_from_world(self, physics_world: PhysicsWorld):
        """
        It is not fully remove, it will be left in memory. if this element is useless in the future, call Func delete()
        """
        if self.origin is not None and self.origin.hasParent():
            self.origin.detachNode()
        self.dynamic_nodes.detach_from_physics_world(physics_world.dynamic_world)
        self.static_nodes.detach_from_physics_world(physics_world.static_world)
        logger.debug("{} is detached from the world.".format(type(self)))

    def destroy(self):
        """
        Fully delete this element and release the memory
        """
        try:
            from metadrive.engine.engine_utils import get_engine
        except ImportError:
            pass
        else:
            engine = get_engine()

            if engine is not None:
                self.detach_from_world(engine.physics_world)
                if self._body is not None and hasattr(self.body, "object"):
                    self.body.generated_object = None
                if self.origin is not None:
                    self.origin.removeNode()

                self.dynamic_nodes.destroy_node_list(bullet_world=engine.physics_world.dynamic_world)
                self.static_nodes.destroy_node_list(bullet_world=engine.physics_world.static_world)

            clear_node_list(self._node_path_list)

            logger.debug("Finish cleaning {} node path.".format(len(self._node_path_list)))
            self._node_path_list.clear()
            self._node_path_list = []

            self.dynamic_nodes.clear()
            self.static_nodes.clear()
            self._config.clear()

    def set_position(self, position, height=None):
        """
        Set this object to a place, the default value is the regular height for red car
        :param position: 2d array or list
        """
        assert len(position) == 2 or len(position) == 3
        if len(position) == 3:
            height = position[-1]
            position = position[:-1]
        else:
            if height is None:
                height = self.origin.getPos()[-1]
        self.origin.setPos(panda_vector(position, height))

    @property
    def position(self):
        return metadrive_vector(self.origin.getPos())

    def set_velocity(self, direction: np.array, value=None, in_local_frame=False):
        """
        Set velocity for object including the direction of velocity and the value (speed)
        The direction of velocity will be normalized automatically, value decided its scale
        :param direction: 2d array or list
        :param value: speed [m/s]
        :param in_local_frame: True, apply speed to local fram
        """
        if in_local_frame:
            direction = self.convert_to_world_coordinates(direction, [0, 0])

        if value is not None:
            norm_ratio = value / (norm(direction[0], direction[1]) + 1e-6)
        else:
            norm_ratio = 1
        self._body.setLinearVelocity(
            LVector3(direction[0] * norm_ratio, direction[1] * norm_ratio,
                     self._body.getLinearVelocity()[-1])
        )

    def set_velocity_km_h(self, direction: list, value=None, in_local_frame=False):
        direction = np.array(direction)
        if value is None:
            direction /= 3.6
        else:
            value /= 3.6
        return self.set_velocity(direction, value, in_local_frame)

    @property
    def velocity(self):
        """
        Velocity, unit: m/s
        """
        velocity = self.body.get_linear_velocity()
        return np.asarray([velocity[0], velocity[1]])

    @property
    def velocity_km_h(self):
        """
        Velocity, unit: m/s
        """
        return self.velocity * 3.6

    @property
    def speed(self):
        """
        return the speed in m/s
        """
        velocity = self.body.get_linear_velocity()
        speed = norm(velocity[0], velocity[1])
        return clip(speed, 0.0, 100000.0)

    @property
    def speed_km_h(self):
        """
        km/h
        """
        velocity = self.body.get_linear_velocity()
        speed = norm(velocity[0], velocity[1]) * 3.6
        return clip(speed, 0.0, 100000.0)

    def set_heading_theta(self, heading_theta, in_rad=True) -> None:
        """
        Set heading theta for this object
        :param heading_theta: float
        :param in_rad: when set to True, heading theta should be in rad, otherwise, in degree
        """
        h = panda_heading(heading_theta)
        if in_rad:
            h = h * 180 / np.pi
        self.origin.setH(h)

    @property
    def heading_theta(self):
        """
        Get the heading theta of this object, unit [rad]
        :return:  heading in rad
        """
        return wrap_to_pi(self.origin.getH() / 180 * math.pi)

    @property
    def heading(self):
        """
        Heading is a vector = [cos(heading_theta), sin(heading_theta)]
        """
        real_heading = self.heading_theta
        # heading = np.array([math.cos(real_heading), math.sin(real_heading)])
        heading = Vector((math.cos(real_heading), math.sin(real_heading)))
        return heading

    @property
    def roll(self):
        """
        Return the roll of this object. As it is facing to x, so roll is pitch
        """
        return np.deg2rad(self.origin.getP())

    def set_roll(self, roll):
        """
        As it is facing to x, so roll is pitch
        """
        self.origin.setP(roll)

    @property
    def pitch(self):
        """
        Return the pitch of this object, as it is facing to x, so pitch is roll
        """
        return np.deg2rad(self.origin.getR())

    def set_pitch(self, pitch):
        """As it is facing to x, so pitch is roll"""
        self.origin.setR(pitch)

    def set_static(self, flag):
        self.body.setStatic(flag)

    def get_panda_pos(self):
        raise DeprecationWarning("It is not allowed to access Panda Pos!")
        return self.origin.getPos()

    def set_panda_pos(self, pos):
        raise DeprecationWarning("It is not allowed to access Panda Pos!")
        self.origin.setPos(pos)

    def get_state(self) -> Dict:
        pos = self.position
        state = {
            ObjectState.POSITION: [pos[0], pos[1], self.get_z()],
            ObjectState.HEADING_THETA: self.heading_theta,
            ObjectState.ROLL: self.roll,
            ObjectState.PITCH: self.pitch,
            ObjectState.VELOCITY: self.velocity,
            ObjectState.TYPE: type(self)
        }
        return state

    def set_state(self, state: Dict):
        self.set_position(state[ObjectState.POSITION])
        self.set_heading_theta(state[ObjectState.HEADING_THETA])
        self.set_pitch(state[ObjectState.PITCH])
        self.set_roll(state[ObjectState.ROLL])
        self.set_velocity(state[ObjectState.VELOCITY])

    @property
    def top_down_color(self):
        rand_c = self.panda_color
        return rand_c[0] * 255, rand_c[1] * 255, rand_c[2] * 255

    @property
    def top_down_width(self):
        raise NotImplementedError(
            "Implement this func for rendering class {} in top down renderer".format(self.class_name)
        )

    @property
    def top_down_length(self):
        raise NotImplementedError(
            "Implement this func for rendering class {} in top down renderer".format(self.class_name)
        )

    def show_coordinates(self):
        pass

    def get_z(self):
        return self.origin.getZ()

    def set_angular_velocity(self, angular_velocity, in_rad=True):
        if not in_rad:
            angular_velocity = angular_velocity / 180 * np.pi
        self._body.setAngularVelocity(LVector3(0, 0, angular_velocity))

    def rename(self, new_name):
        super(BaseObject, self).rename(new_name)
        physics_node = self._body.getPythonTag(self._body.getName())
        if isinstance(physics_node, BaseGhostBodyNode) or isinstance(physics_node, BaseRigidBodyNode):
            physics_node.rename(new_name)

    def random_rename(self):
        self.rename(random_string())

    def convert_to_local_coordinates(self, vector, origin):
        """
        Give vector in world coordinates, and convert it to object coordinates. For example, vector can be other vehicle
        position, origin could be this vehicles position. In this case, vector-origin will be transformed to ego car
        coordinates. If origin is set to 0, then no offset is applied and this API only calculates relative direction.

        In a word, for calculating **points transformation** in different coordinates, origin is required. This is
        because vectors have no origin but origin is required to define a point.
        """
        vector = np.asarray(vector) - np.asarray(origin)
        vector = LVector3(*vector, 0.)
        vector = self.origin.getRelativeVector(self.engine.origin, vector)
        project_on_x = vector[0]
        project_on_y = vector[1]
        return np.array([project_on_x, project_on_y])

    def convert_to_world_coordinates(self, vector, origin):
        """
        Give a vector in local coordinates, and convert it to world coordinates. The origin should be added as offset.
        For example, vector could be a relative position in local coordinates and origin could be ego car's position.
        If origin is set to 0, then no offset is applied and this API only calculates relative direction.

        In a word, for calculating **points transformation** in different coordinates, origin is required. This is
        because vectors have no origin but origin is required to define a point.
        """
        assert len(vector) == 2 or len(vector) == 3, "the vector should be in shape (2,) or (3,)"
        vector = vector[:2]
        vector = LVector3(*vector, 0.)
        vector = self.engine.origin.getRelativeVector(self.origin, vector)
        project_on_x = vector[0]
        project_on_y = vector[1]
        return np.array([project_on_x, project_on_y]) + np.asarray(origin)

    @property
    def WIDTH(self):
        raise NotImplementedError()

    @property
    def LENGTH(self):
        raise NotImplementedError()

    @property
    def bounding_box(self):
        """
        This function will return the 2D bounding box of vehicle. Points are in clockwise sequence, first point is the
        top-left point.
        """
        p1 = self.convert_to_world_coordinates([self.LENGTH / 2, self.WIDTH / 2], self.position)
        p2 = self.convert_to_world_coordinates([self.LENGTH / 2, -self.WIDTH / 2], self.position)
        p3 = self.convert_to_world_coordinates([-self.LENGTH / 2, -self.WIDTH / 2], self.position)
        p4 = self.convert_to_world_coordinates([-self.LENGTH / 2, self.WIDTH / 2], self.position)
        return [p1, p2, p3, p4]

    @property
    def use_render_pipeline(self):
        return self.engine is not None and self.engine.use_render_pipeline
